@@ -56,11 +56,32 @@ pub trait SOP {
 }
 
 pub trait Version<'a> {
-    /// Returns the implementation's name.
-    fn name(&self) -> String;
+    /// Returns name and version of the SOP implementation.
+    fn frontend(&self) -> Result<VersionInfo>;
 
-    /// Returns the implementation's version string.
-    fn version(&self) -> String;
+    /// Returns name and version of the primary underlying OpenPGP
+    /// toolkit.
+    fn backend(&self) -> Result<VersionInfo>;
+
+    /// Returns multiple name and version pairs.
+    ///
+    /// The first one MUST match the information produced by
+    /// [`Version::frontend`], but the rest have no defined structure.
+    /// This is what the default implementation does.
+    fn extended(&self) -> Result<Vec<VersionInfo>> {
+        Ok(vec![
+            self.frontend()?,
+            self.backend()?,
+        ])
+    }
+}
+
+/// Represents a name and version tuple.
+pub struct VersionInfo {
+    /// Name of the implementation, library, or additional component.
+    pub name: String,
+    /// Version string.
+    pub version: String,
 }
 
 /// Builder for [`SOP::generate_key`].
@@ -83,6 +104,10 @@ pub trait ExtractCert<'a> {
     /// Extracts the cert from `key`.
     fn key(self: Box<Self>, key: &mut (dyn io::Read + Send + Sync))
            -> Result<Box<dyn Ready + 'a>>;
+
+    /// Extracts the certs from `keys`.
+    fn keys(self: Box<Self>, keys: &mut (dyn io::Read + Send + Sync))
+            -> Result<Box<dyn Ready + 'a>>;
 }
 
 /// Builder for [`SOP::sign`].
@@ -105,7 +130,7 @@ pub trait Sign<'a> {
 
     /// Signs data.
     fn data(self: Box<Self>, data: &'a mut (dyn io::Read + Send + Sync))
-        -> Result<Box<dyn Ready + 'a>>;
+            -> Result<Box<dyn ReadyWithResult<Micalg> + 'a>>;
 }
 
 /// Builder for [`SOP::verify`].
@@ -290,6 +315,10 @@ pub trait Decrypt<'a> {
     fn with_key(self: Box<Self>, key: &mut (dyn io::Read + Send + Sync))
                 -> Result<Box<dyn Decrypt<'a> + 'a>>;
 
+    /// Adds the decryption keys.
+    fn with_keys(self: Box<Self>, key: &mut (dyn io::Read + Send + Sync))
+                 -> Result<Box<dyn Decrypt<'a> + 'a>>;
+
     /// Decrypts `ciphertext`, returning verification results and
     /// plaintext.
     fn ciphertext(self: Box<Self>,
@@ -665,6 +694,63 @@ impl fmt::Display for ArmorLabel {
     }
 }
 
+/// Indicates the cryptographic digest used when making a signature.
+///
+/// It is useful specifically when generating signed PGP/MIME objects,
+/// which want a `micalg=` parameter for the `multipart/signed`
+/// content type as described in section 5 of [RFC3156].
+///
+/// [RFC3156]: https://datatracker.ietf.org/doc/html/rfc3156
+pub enum Micalg {
+    /// Rivest et.al. message digest 5 (deprecated).
+    MD5,
+    /// NIST Secure Hash Algorithm (deprecated).
+    SHA1,
+    /// RIPEMD-160 (deprecated).
+    RipeMD,
+    /// 256-bit version of SHA2.
+    SHA256,
+    /// 384-bit version of SHA2.
+    SHA384,
+    /// 512-bit version of SHA2.
+    SHA512,
+    /// 224-bit version of SHA2.
+    SHA224,
+    /// Unknown hash algorithm.
+    Unknown(String),
+}
+
+impl From<u8> for Micalg {
+    fn from(o: u8) -> Self {
+        match o {
+            1 => Micalg::MD5,
+            2 => Micalg::SHA1,
+            3 => Micalg::RipeMD,
+            8 => Micalg::SHA256,
+            9 => Micalg::SHA384,
+            10 => Micalg::SHA512,
+            11 => Micalg::SHA224,
+            u => Micalg::Unknown(format!("unknown-algo-{}", u)),
+        }
+    }
+}
+
+impl fmt::Display for Micalg {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.write_str("pgp-")?;
+        match self {
+            Micalg::MD5 => f.write_str("md5"),
+            Micalg::SHA1 => f.write_str("sha1"),
+            Micalg::RipeMD => f.write_str("ripemd160"),
+            Micalg::SHA256 => f.write_str("sha256"),
+            Micalg::SHA384 => f.write_str("sha384"),
+            Micalg::SHA512 => f.write_str("sha512"),
+            Micalg::SHA224 => f.write_str("sha224"),
+            Micalg::Unknown(a) => f.write_str(&a.to_lowercase()),
+        }
+    }
+}
+
 /// Result specialization.
 pub type Result<T> = std::result::Result<T, Error>;
 
@@ -688,10 +774,11 @@ pub enum Error {
     #[error("Certificate not encryption-capable")]
     CertCannotEncrypt,
 
-    /// Certificate not signing-capable (e.g., expired, revoked,
-    /// unacceptable usage flags) ("sop sign").
-    #[error("Certificate not signing-capable")]
-    CertCannotSign,
+    /// Key not signature-capable (e.g., expired, revoked,
+    /// unacceptable usage flags) (`sop sign` and `sop encrypt` with
+    /// `--sign-with`).
+    #[error("Key not signing-capable")]
+    KeyCannotSign,
 
     /// Missing required argument.
     #[error("Missing required argument")]
